//
// PWinObj.hh for pekwm
// Copyright (C)  2003-2004 Claes Nasten <pekdon{@}pekdon{.}net>
//
// This program is licensed under the GNU GPL.
// See the LICENSE file for more information.
//

#include "../config.h"

#ifndef _WINDOW_OBJECT_HH_
#define _WINDOW_OBJECT_HH_

#include "pekwm.hh"
#include "Action.hh"

#include <list>
#include <algorithm>

class PWinObj
{
public:
	enum Type {
		WO_FRAME = (1<<1),
		WO_CLIENT = (1<<2),
		WO_MENU = (1<<3),
		WO_DOCKAPP = (1<<4),
		WO_SCREEN_EDGE = (1<<5),
		WO_SCREEN_ROOT = (1<<6),
		WO_CMD_DIALOG = (1<<7),
		WO_STATUS = (1<<8),
		WO_NO_TYPE = 0
	};

	PWinObj(Display *dpy);
	virtual ~PWinObj(void);

	static inline PWinObj *getFocusedPWinObj(void) {
		return _focused_wo;
	}
	static inline PWinObj *getRootPWinObj(void) {
		return _root_wo;
	}
	static inline void setFocusedPWinObj(PWinObj *wo) {
		_focused_wo = wo;
	}
	static inline void setRootPWinObj(PWinObj *wo) {
		_root_wo = wo;
	}

	static inline PWinObj *findPWinObj(Window win) {
		std::list<PWinObj*>::iterator it = _wo_list.begin();

		for (; it != _wo_list.end(); ++it) {
			if (*(*it) == win) {
				return *it;
			}
		}

		return NULL;
	}
	static inline bool windowObjectExists(PWinObj *wo) {
		std::list<PWinObj*>::iterator it =
			find(_wo_list.begin(), _wo_list.end(), wo);
		if (it != _wo_list.end())
			return true;
		return false;
	}

	inline Window getWindow(void) const { return _window; }
	inline void setWindow(Window window) { _window = window; }
	inline PWinObj *getParent(void) const { return _parent; }
	inline void setParent(PWinObj *wo) { _parent = wo; }
	inline Type getType(void) const { return _type; }

	inline int getX(void) const { return _gm.x; }
	inline int getY(void) const { return _gm.y; }
	inline int getRX(void) const { return _gm.x + _gm.width; }
	inline int getBY(void) const { return _gm.y + _gm.height; }

	inline uint getWidth(void) const { return _gm.width; }
	inline uint getHeight(void) const { return _gm.height; }

	inline uint getWorkspace(void) const { return _workspace; }
	inline uint getLayer(void) const { return _layer; }

	inline bool isMapped(void) const { return _mapped; }
	inline bool isIconified(void) const { return _iconified; }
	inline bool isHidden(void) const { return _hidden; }
	inline bool isFocused(void) const { return _focused; }
	inline bool isSticky(void) const { return _sticky; }

	inline bool isFocusable(void) const { return _focusable; }

	// interface
	virtual void mapWindow(void);
	virtual void mapWindowRaised(void);
	virtual void unmapWindow(void);
	virtual void iconify(void);
	virtual void stick(void);

	virtual void move(int x, int y, bool do_virtual = true);
	virtual void moveVirtual(int x, int y);
	virtual void resize(uint width, uint height);
	virtual void raise(void) { XRaiseWindow(_dpy, _window); }
	virtual void lower(void) { XLowerWindow(_dpy, _window); }

	virtual void setWorkspace(uint workspace);
	virtual void setLayer(uint layer);
	virtual void setFocused(bool focused);
	virtual void setSticky(bool sticky);
	virtual void setHidden(bool hidden);

	virtual bool giveInputFocus(void);
	virtual void reparent(PWinObj *parent, int x, int y);

	// event interface
	virtual ActionEvent *handleButtonPress(XButtonEvent *ev) { return NULL; }
	virtual ActionEvent *handleButtonRelease(XButtonEvent *ev) { return NULL; }
	virtual ActionEvent *handleKeyPress(XKeyEvent *ev) { return NULL; }
	virtual ActionEvent *handleKeyRelease(XKeyEvent *ev) { return NULL; }
	virtual ActionEvent *handleMotionEvent(XMotionEvent *ev) { return NULL; } 
	virtual ActionEvent *handleEnterEvent(XCrossingEvent *ev) { return NULL; }
	virtual ActionEvent *handleLeaveEvent(XCrossingEvent *ev) { return NULL; }
	virtual ActionEvent *handleExposeEvent(XExposeEvent *ev) { return NULL; }

	virtual ActionEvent *handleMapRequest(XMapRequestEvent *ev) { return NULL; }
	virtual ActionEvent *handleUnmapEvent(XUnmapEvent *ev) { return NULL; }

	// operators
	virtual bool operator == (const Window &window) {
		return (_window == window);
	}
	virtual bool operator != (const Window &window) {
		return (_window != window);
	}

	// other window commands
	inline void clear(void) { XClearWindow(_dpy, _window); }
	inline void setBackground(long pixel) { XSetWindowBackground(_dpy, _window, pixel); }
	inline void setBackgroundPixmap(Pixmap pm) { XSetWindowBackgroundPixmap(_dpy, _window, pm); }

protected:
	Display *_dpy;
	Window _window;
	PWinObj *_parent;

	Type _type;

	Geometry _gm;
	int _v_x, _v_y;
	uint _workspace, _layer;
	bool _mapped, _iconified, _hidden, _focused, _sticky;
	bool _focusable;

	static PWinObj *_root_wo, *_focused_wo;
	static std::list<PWinObj*> _wo_list;
};

#endif // _WINDOW_OBJECT_HH_
