//
// Theme.cc for pekwm
// Copyright (C) 2003-2004 Claes Nasten <pekdon{@}pekdon{.}net>
//
// This program is licensed under the GNU GPL.
// See the LICENSE file for more information.
//

#include "../config.h"
#include "ParseUtil.hh"
#include "Theme.hh"

#include "PScreen.hh"
#include "Config.hh"
#include "PWinObj.hh"
#include "PFont.hh"
#include "PTexture.hh"
#include "Image.hh"
#include "ColorHandler.hh"
#include "FontHandler.hh"
#include "ImageHandler.hh"
#include "TextureHandler.hh"
#include "Util.hh"

#include <iostream>
#include <cstdlib>

using std::cerr;
using std::endl;
using std::string;
using std::list;
using std::vector;
using std::map;

// Theme::PDecorButtonData

//! @brief
Theme::PDecorButtonData::PDecorButtonData(void) :
_left(false), _width(1), _height(1)
{
	for (uint i = 0; i < BUTTON_STATE_NO; ++i) {
		_texture[i] = NULL;
	}
}

//! @brief
Theme::PDecorButtonData::~PDecorButtonData(void)
{
	unload();
}

//! @brief
bool
Theme::PDecorButtonData::load(CfgParser::Section *cs)
{
	if (*cs == "LEFT") {
		_left = true;
	} else if (*cs == "RIGHT") {
		_left = false;
	} else {
		return false; // not a button section
	}

	// get actions
	ActionEvent ae;
	CfgParser::Section *sub;
	while ((sub = cs->getNextSection()) != NULL) {
		if (Config::instance()->parseActionEvent(sub, ae, BUTTONCLICK_OK, true)) {
			_ae_list.push_back(ae);
		}
	}

	// got some actions, consider it to be a valid button
	if (_ae_list.size() > 0) {
		string s_val;

		if (cs->getValue("FOCUSED", s_val)) {
			_texture[BUTTON_STATE_FOCUSED] = TextureHandler::instance()->getTexture(s_val);
		}
		if (cs->getValue("UNFOCUSED", s_val)) {
			_texture[BUTTON_STATE_UNFOCUSED] = TextureHandler::instance()->getTexture(s_val);
		}
		if (cs->getValue("PRESSED", s_val)) {
			_texture[BUTTON_STATE_PRESSED] = TextureHandler::instance()->getTexture(s_val);
		}
		if (cs->getValue("HOOVER", s_val)) {
			_texture[BUTTON_STATE_HOOVER] = TextureHandler::instance()->getTexture(s_val);
		}

		check();

		_width = _texture[BUTTON_STATE_FOCUSED]->getWidth();
		_height = _texture[BUTTON_STATE_FOCUSED]->getHeight();
		if (_width == 0) {
			cerr << " *** WARNING: button texture width == 0" << endl;
			_width = 1;
		}
		if (_height == 0) {
			cerr << " *** WARNING: button texture height == 0" << endl;
			_height = 1;
		}

		return true;
	}

	return false;
}

//! @brief
void
Theme::PDecorButtonData::unload(void)
{
	for (uint i = 0; i < BUTTON_STATE_NO; ++i) {
		TextureHandler::instance()->returnTexture(_texture[i]);
		_texture[i] = NULL;
	}
}

//! @brief
void
Theme::PDecorButtonData::check(void)
{
	for (uint i = 0; i < BUTTON_STATE_NO; ++i) {
		if (_texture[i] == NULL) {
			_texture[i] = TextureHandler::instance()->getTexture("EMPTY");
		}
	}
}

// Theme::PDecorData

map<FocusedState, string> Theme::PDecorData::_fs_map = map<FocusedState, string>();
map<BorderPosition, string> Theme::PDecorData::_border_map = map<BorderPosition, string>();

//! @brief
Theme::PDecorData::PDecorData(void) :
_title_height(0)
{
	// init static data
	if (_fs_map.size() == 0) {
		_fs_map[FOCUSED_STATE_FOCUSED] = "FOCUSED";
		_fs_map[FOCUSED_STATE_UNFOCUSED] = "UNFOCUSED";
		_fs_map[FOCUSED_STATE_FOCUSED_SELECTED] = "FOCUSEDSELECTED";
		_fs_map[FOCUSED_STATE_UNFOCUSED_SELECTED] = "UNFOCUSEDSELECTED";
	}
	if (_border_map.size() == 0) {
		_border_map[BORDER_TOP_LEFT] = "TOPLEFT";
		_border_map[BORDER_TOP] = "TOP";
		_border_map[BORDER_TOP_RIGHT] = "TOPRIGHT";
		_border_map[BORDER_LEFT] = "LEFT";
		_border_map[BORDER_RIGHT] = "RIGHT";
		_border_map[BORDER_BOTTOM_LEFT] = "BOTTOMLEFT";
		_border_map[BORDER_BOTTOM] = "BOTTOM";
		_border_map[BORDER_BOTTOM_RIGHT] = "BOTTOMRIGHT";
	}

	for (uint i = 0; i < DIRECTION_NO; ++i) {
		_pad[i] = 0;
	}
	for (uint i = 0; i < FOCUSED_STATE_NO; ++i) {
		_texture_tab[i] = NULL;
		_font[i] = NULL;
		_font_color[i] = NULL;
	}
	for (uint i = 0; i < FOCUSED_STATE_FOCUSED_SELECTED; ++i) {
		_texture_main[i] = NULL;
		_texture_separator[i] = NULL;
	}
}

//! @brief
Theme::PDecorData::~PDecorData(void)
{
	unload();
}

//! @brief
bool
Theme::PDecorData::load(CfgParser::Section *cs)
{
	string s_val;
	cs->getValue("NAME", s_val);

	if (s_val == "") {
		cerr << " *** WARNING: no name identifying decor" << endl;
		return false;
	}
	_name = s_val;

	TextureHandler *th = TextureHandler::instance(); // convenience

	CfgParser::Section *sub, *sub_2;

	if ((sub = cs->getSection("TITLE")) != NULL) {
		sub->getValue("HEIGHT", _title_height);
		if (sub->getValue("PAD", s_val)) {
			vector<string> tok;
			if (Util::splitString(s_val, tok, " \t", 4) == 4) {
				for (uint i = 0; i < DIRECTION_NO; ++i) {
					_pad[i] = strtol(tok[i].c_str(), NULL, 10);
				}
			}
		}
		if (sub->getValue("FOCUSED", s_val)) {
			_texture_main[FOCUSED_STATE_FOCUSED] = th->getTexture(s_val);
		}
		if (sub->getValue("UNFOCUSED", s_val)) {
			_texture_main[FOCUSED_STATE_UNFOCUSED] = th->getTexture(s_val);
		}
		if ((sub_2 = sub->getSection("TAB")) != NULL) {
			for (uint i = 0; i < FOCUSED_STATE_NO; ++i) {
				if (sub_2->getValue(_fs_map[FocusedState(i)], s_val) == true) {
					_texture_tab[i] = th->getTexture(s_val);
				}
			}
		}
		if ((sub_2 = sub->getSection("SEPARATOR")) != NULL) {
			if (sub_2->getValue("FOCUSED", s_val)) {
				_texture_separator[FOCUSED_STATE_FOCUSED] = th->getTexture(s_val);
			}
			if (sub_2->getValue("UNFOCUSED", s_val)) {
				_texture_separator[FOCUSED_STATE_UNFOCUSED] = th->getTexture(s_val);
			}
		}
		if ((sub_2 = sub->getSection("FONT")) != NULL) {
			// font
			for (uint i = 0; i < FOCUSED_STATE_NO; ++i) {
				if (sub_2->getValue(_fs_map[FocusedState(i)], s_val) == true) {
					_font[i] = FontHandler::instance()->getFont(s_val);
				}
			}
		} else {
			cerr << " *** WARNING: no font section in decor: " << _name << endl;
		}
		if ((sub_2 = sub->getSection("FONTCOLOR")) != NULL) {
			// font color
			for (uint i = 0; i < FOCUSED_STATE_NO; ++i) {
				if (sub_2->getValue(_fs_map[FocusedState(i)], s_val) == true) {
					_font_color[i] = FontHandler::instance()->getColor(s_val);
				}
			}
		}
		if ((sub_2 = sub->getSection("BUTTONS")) != NULL) {
			// buttons
			loadButtons(sub_2);
		}
		if ((sub_2 = sub->getSection("BORDER")) != NULL) {
			// border
			loadBorder(sub_2);
		}
	} else {
		cerr << " *** WARNING: no title section in decor: " << _name << endl;
	}
	// properties

	check();

	return true;
}

//! @brief
void
Theme::PDecorData::unload(void)
{
	TextureHandler *th = TextureHandler::instance();

	for (uint i = 0; i < FOCUSED_STATE_NO; ++i) {
		th->returnTexture(_texture_tab[i]);
		FontHandler::instance()->returnFont(_font[i]);
		FontHandler::instance()->returnColor(_font_color[i]);

		_texture_tab[i] = NULL;
		_font[i] = NULL;
		_font_color[i] = NULL;
	}

	for (uint i = 0; i < FOCUSED_STATE_FOCUSED_SELECTED; ++i) {
		th->returnTexture(_texture_main[i]);
		th->returnTexture(_texture_separator[i]);
		_texture_main[i] = NULL;
		_texture_separator[i] = NULL;

		for (uint j = 0; j < BORDER_NO_POS; ++j) {
			th->returnTexture(_texture_border[i][j]);
			_texture_border[i][j] = NULL;
		}
	}

	list<Theme::PDecorButtonData*>::iterator it(_button_list.begin());
	for (; it != _button_list.end(); ++it) {
		delete *it;
	}
	_button_list.clear();
}

//! @brief Checks data properties, prints warning and tries to fix
void
Theme::PDecorData::check(void)
{
	checkTextures();
	checkFonts();
	checkBorder();
	checkColors();
}

//! @brief
void
Theme::PDecorData::loadBorder(CfgParser::Section *cs)
{
	TextureHandler *th = TextureHandler::instance(); // convenience

	string s_val;
	CfgParser::Section *sub;

	if ((sub = cs->getSection("FOCUSED")) != NULL) {
		for (uint i = 0; i < BORDER_NO_POS; ++i) {
			if (sub->getValue(_border_map[BorderPosition(i)], s_val)) {
				_texture_border[FOCUSED_STATE_FOCUSED][i] = th->getTexture(s_val);
			}
		}
	}
	if ((sub = cs->getSection("UNFOCUSED")) != NULL) {
		for (uint i = 0; i < BORDER_NO_POS; ++i) {
			if (sub->getValue(_border_map[BorderPosition(i)], s_val)) {
				_texture_border[FOCUSED_STATE_UNFOCUSED][i] = th->getTexture(s_val);
			}
		}
	}
}

//! @brief
void
Theme::PDecorData::loadButtons(CfgParser::Section *cs)
{
	if (cs == NULL) {
		return;
	}

	CfgParser::Section *sub;
	while ((sub = cs->getNextSection()) != NULL) {
		Theme::PDecorButtonData *btn = new Theme::PDecorButtonData();
		if (btn->load(sub) == true) {
			_button_list.push_back(btn);
		} else {
			delete btn;
		}
	}
}

//! @brief Checks for NULL textures, prints warning and sets empty texture
void
Theme::PDecorData::checkTextures(void)
{
	for (uint i = 0; i < FOCUSED_STATE_NO; ++i) {
		if (_texture_tab[i] == NULL) {
			cerr << " *** WARNING: " << _name << " missing tab texture state "
					 << _fs_map[FocusedState(i)] << endl;
			_texture_tab[i] = TextureHandler::instance()->getTexture("EMPTY");
		}
	}
	for (uint i = 0; i < FOCUSED_STATE_FOCUSED_SELECTED; ++i) {
		if (_texture_main[i] == NULL) {
			cerr << " *** WARNING: " << _name << " missing main texture state "
					 << _fs_map[FocusedState(i)] << endl;
			_texture_main[i] = TextureHandler::instance()->getTexture("EMPTY");
		}
		if (_texture_separator[i] == NULL) {
			cerr << " *** WARNING: " << _name << " missing tab texture state "
					 << _fs_map[FocusedState(i)] << endl;
			_texture_separator[i] = TextureHandler::instance()->getTexture("EMPTY");
		}
	}
}

//! @brief Checks for NULL fonts, prints warning and sets empty font
void
Theme::PDecorData::checkFonts(void)
{
	// the only font that's "obligatory" is the standard focused font,
	// others are only used if availible so we only check the focused font.
	if (_font[FOCUSED_STATE_FOCUSED] == NULL) {
		cerr << " *** WARNING: " << _name << " missing font state "
				 << _fs_map[FOCUSED_STATE_FOCUSED] << endl;
		_font[FOCUSED_STATE_FOCUSED] = FontHandler::instance()->getFont("");
	}
}

//! @brief
void
Theme::PDecorData::checkBorder(void)
{
	for (uint state = FOCUSED_STATE_FOCUSED; state < FOCUSED_STATE_FOCUSED_SELECTED; ++state) {
		for (uint i = 0; i < BORDER_NO_POS; ++i) {
			if (_texture_border[state][i] == NULL) {
				cerr << " *** WARNING: border texture " << _border_map[BorderPosition(i)]
						 << " " << _fs_map[FocusedState(state)] << " missing" << endl;
				_texture_border[state][i] = TextureHandler::instance()->getTexture("EMPTY");
			}
		}
	}
}

//! @brief Checks for NULL colors, prints warning and sets empty color
void
Theme::PDecorData::checkColors(void)
{
	for (uint i = 0; i < FOCUSED_STATE_NO; ++i) {
		if (_font_color[i] == NULL) {
			cerr << " *** WARNING: " << _name << " missing font color state "
					 << _fs_map[FocusedState(i)] << endl;
			_font_color[i] = FontHandler::instance()->getColor("#000000");
		}
	}
}

// Theme::PMenuData

//! @brief PMenuData constructor
Theme::PMenuData::PMenuData(void)
{
	for (uint i = 0; i <= OBJECT_STATE_NO; ++i) {
		_tex_menu[i] = NULL;
		_tex_item[i] = NULL;
		_tex_arrow[i] = NULL;
	}
	for (uint i = 0; i < OBJECT_STATE_NO; ++i) {
		_tex_sep[i] = NULL;
	}
	for (uint i = 0; i < DIRECTION_NO; ++i) {
		_pad[i] = 0;
	}
}

//! @brief PMenuData destructor
Theme::PMenuData::~PMenuData(void)
{
	unload();
}

//! @brief
void
Theme::PMenuData::load(CfgParser::Section *cs)
{
	string s_val;
	CfgParser::Section *sub;

	if (cs->getValue("PAD", s_val)) {
		vector<string> tok;
		if (Util::splitString(s_val, tok, " \t", 4) == 4) {
			for (int i = 0; i < DIRECTION_NO; ++i) {
				_pad[i] = strtol(tok[i].c_str(), NULL, 10);
			}
		}
	}

	if ((sub = cs->getSection("FOCUSED")) != NULL) {
		loadState(sub, OBJECT_STATE_FOCUSED);
	}
	if ((sub = cs->getSection("UNFOCUSED")) != NULL) {
		loadState(sub, OBJECT_STATE_UNFOCUSED);
	}
	if ((sub = cs->getSection("SELECTED")) != NULL) {
		loadState(sub, OBJECT_STATE_SELECTED);
	}

	check();
}

//! @brief
//! @todo print warning
void
Theme::PMenuData::unload(void)
{
	for (uint i = 0; i <= OBJECT_STATE_NO; ++i) {
		FontHandler::instance()->returnFont(_font[i]);
		FontHandler::instance()->returnColor(_color[i]);
		TextureHandler::instance()->returnTexture(_tex_menu[i]);
		TextureHandler::instance()->returnTexture(_tex_item[i]);
		TextureHandler::instance()->returnTexture(_tex_arrow[i]);

		_font[i] = NULL;
		_color[i] = NULL;
		_tex_menu[i] = NULL;
		_tex_item[i] = NULL;
		_tex_arrow[i] = NULL;
	}

	for (uint i = 0; i < OBJECT_STATE_NO; ++i) {
		TextureHandler::instance()->returnTexture(_tex_sep[i]);
		_tex_sep[i] = NULL;
	}
}

//! @brief
void
Theme::PMenuData::check(void)
{
	for (uint i = 0; i <= OBJECT_STATE_NO; ++i) {
		if (_font[i] == NULL) {
			_font[i] = FontHandler::instance()->getFont("");
		}
		if (_color[i] == NULL) {
			_color[i] = FontHandler::instance()->getColor("#000000");
		}
		if (_tex_menu[i] == NULL) {
			_tex_menu[i] = TextureHandler::instance()->getTexture("EMPTY");
		}
		if (_tex_item[i] == NULL) {
			_tex_item[i] = TextureHandler::instance()->getTexture("EMPTY");
		}
		if (_tex_arrow[i] == NULL) {
			_tex_arrow[i] = TextureHandler::instance()->getTexture("EMPTY");
		}
	}

	for (uint i = 0; i < OBJECT_STATE_NO; ++i) {
		if (_tex_sep[i] == NULL) {
			_tex_sep[i] = TextureHandler::instance()->getTexture("EMPTY");
		}
	}
}

//! @brief
void
Theme::PMenuData::loadState(CfgParser::Section *cs, ObjectState state)
{
	string s_val;
	TextureHandler *th = TextureHandler::instance();

	if (cs->getValue("FONT", s_val)) {
		_font[state] = FontHandler::instance()->getFont(s_val);
	}
	if (cs->getValue("BACKGROUND", s_val)) {
		_tex_menu[state] = th->getTexture(s_val);
	}
	if (cs->getValue("ITEM", s_val)) {
		_tex_item[state] = th->getTexture(s_val);
	}
	if (cs->getValue("TEXT", s_val)) {
		_color[state] = FontHandler::instance()->getColor(s_val);
	}
	if (cs->getValue("ARROW", s_val)) {
		_tex_arrow[state] = th->getTexture(s_val);
	}

	if (state < OBJECT_STATE_SELECTED) {
		if (cs->getValue("SEPARATOR", s_val)) {
			_tex_sep[state] = th->getTexture(s_val);
		}
	}
}

// Theme::TextDialogData

//! @brief TextDialogData constructor
Theme::TextDialogData::TextDialogData(void) :
_font(NULL), _color(NULL), _tex(NULL)
{
	for (uint i = 0; i < DIRECTION_NO; ++i) {
		_pad[i] = 0;
	}
}

//! @brief
Theme::TextDialogData::~TextDialogData(void)
{
	unload();
}
//! @brief
void
Theme::TextDialogData::load(CfgParser::Section *cs)
{
	string s_val;

	if (cs->getValue("FONT", s_val)) {
		_font = FontHandler::instance()->getFont(s_val);
	}
	if (cs->getValue("TEXT", s_val)) {
		_color = FontHandler::instance()->getColor(s_val);
	}
	if (cs->getValue("TEXTURE", s_val)) {
		_tex = TextureHandler::instance()->getTexture(s_val);
	}
	if (cs->getValue("PAD", s_val)) {
		vector<string> tok;
		if (Util::splitString(s_val, tok, " \t", 4) == 4) {
			for (uint i = 0; i < DIRECTION_NO; ++i) {
				_pad[i] = strtol(tok[i].c_str(), NULL, 10);
			}
		}
	}
}

//! @brief
void
Theme::TextDialogData::unload(void)
{
	FontHandler::instance()->returnFont(_font);
	FontHandler::instance()->returnColor(_color);
	TextureHandler::instance()->returnTexture(_tex);

	_font = NULL;
	_tex = NULL;
	_color = NULL;
}

//! @brief
//! @todo print warnings
void
Theme::TextDialogData::check(void)
{
	if (_font == NULL) {
		_font = FontHandler::instance()->getFont("");
	}
	if (_color == NULL) {
		_color = FontHandler::instance()->getColor("#000000");
	}
	if (_tex == NULL) {
		_tex = TextureHandler::instance()->getTexture("EMPTY");
	}
}

// Theme

//! @brief Theme constructor
Theme::Theme(PScreen *scr) :
_scr(scr),
_is_loaded(false),
_invert_gc(None)
{
	// image handler
	_image_handler = new ImageHandler();

	// window gc's
	XGCValues gv;
	gv.function = GXinvert;
	gv.subwindow_mode = IncludeInferiors;
	gv.line_width = 1;
	_invert_gc = XCreateGC(_scr->getDpy(), _scr->getRoot(),
												 GCFunction|GCSubwindowMode|GCLineWidth, &gv);


	_scr->grabServer();

	load(Config::instance()->getThemeFile());

	_scr->ungrabServer(true);
}

//! @brief Theme destructor
Theme::~Theme(void)
{
	unload(); // should clean things up

	XFreeGC(_scr->getDpy(), _invert_gc);

	delete _image_handler;
}

//! @brief Loads the "ThemeFile", unloads any previous loaded theme.
void
Theme::load(const std::string &dir)
{
	if (_is_loaded)
		unload();

	CfgParser theme;
	CfgParser::Section *cs, *sub;

	_theme_dir = dir;
	if (_theme_dir.at(_theme_dir.size() - 1) != '/')
		_theme_dir.append("/");

	string theme_file = _theme_dir + string("theme");
	if (theme.load(theme_file) == false) {
		theme_file = DATADIR "/themes/default";
		if (theme.load(theme_file) == false) {
			cerr << " *** ERROR: couldn't load theme: " << _theme_dir
					 << " or default theme." << endl;
			exit(1);
		}
	}

	// set image basedir
	_image_handler->setDir(_theme_dir);

	// temporary variables used in parsing
	string s_val;

	if ((cs = theme.getSection("PDECOR")) != NULL) {
		while ((sub = cs->getNextSection()) != NULL) {
			Theme::PDecorData *data = new Theme::PDecorData();

			if (data->load(sub) == true) {
				_pdecordata_map[data->getName()] = data;
			} else {
				delete data;
			}
		}
	}

	if ((cs = theme.getSection("MENU"))) {
		_menu_data.load(cs);
	}

	if ((cs = theme.getSection("ROOT"))) {
		if (cs->getValue("COMMAND", s_val)) {
			Util::forkExec(s_val);
		}
	}


#ifdef HARBOUR
	if ((cs = theme.getSection("HARBOUR"))) {
		if (cs->getValue("TEXTURE", s_val)) {
			_harbour_texture = TextureHandler::instance()->getTexture(s_val);
		}
	}
#endif // HARBOUR

	// status window
	if ((cs = theme.getSection("STATUS")) != NULL) {
		_status_data.load(cs);
	} else {
		_status_data.check();
	}
	// cmd window
	if ((cs = theme.getSection("CMDDIALOG")) != NULL) {
		_cmd_d_data.load(cs);
	} else {
		_cmd_d_data.check();
	}

	_is_loaded = true;

	// validate theme
	if (getPDecorData("DEFAULT") == NULL) {
		cerr << " *** ERROR: theme doesn't contain any DEFAULT decor." << endl;
		exit(1);
	}
}

//! @brief Unloads all pixmaps, fonts, gc and colors allocated by the theme.
void
Theme::unload(void) {
	map<string, Theme::PDecorData*>::iterator it(_pdecordata_map.begin());
	for (; it != _pdecordata_map.end(); ++it) {
		delete it->second;
	}
	_pdecordata_map.clear();

#ifdef HARBOUR
	TextureHandler::instance()->returnTexture(_harbour_texture);
#endif // HARBOUR

	// unload menu
	_menu_data.unload();

	_status_data.unload();
	_cmd_d_data.unload();

	_is_loaded = false;
}
