//
// Theme.hh for pekwm
// Copyright (C) 2003-2004 Claes Nasten <pekdon{@}pekdon{.}net>
//
// This program is licensed under the GNU GPL.
// See the LICENSE file for more information.
//

#include "../config.h"

#ifndef _THEME_HH_
#define _THEME_HH_

#include "pekwm.hh"
#include "CfgParser.hh"
#include "Action.hh" // ActionEvent
#include "PFont.hh" // PFont::Color
#include "ParseUtil.hh"

class PScreen;
class PTexture;
class Button;
class ButtonData;
class Image;
class ImageHandler;

#include <string>
#include <map>

class Theme
{
public:
	class PDecorButtonData {
	public:
		PDecorButtonData(void);
		~PDecorButtonData(void);

		inline bool isLeft(void) const { return _left; }
		inline uint getWidth(void) const { return _width; }
		inline uint getHeight(void) const { return _height; }

		inline PTexture *getTexture(ButtonState state) { if (state != BUTTON_STATE_NO) { return _texture[state]; } return NULL; }
		inline std::list<ActionEvent>::iterator begin(void) { return _ae_list.begin(); }
		inline std::list<ActionEvent>::iterator end(void) { return _ae_list.end(); }

		bool load(CfgParser::Section *cs);
		void unload(void);

	private:
		void check(void);

	private:
		std::list<ActionEvent> _ae_list;
		PTexture *_texture[BUTTON_STATE_NO];

		bool _left;
		uint _width, _height;
	};

	class PDecorData {
	public:
		PDecorData(void);
		~PDecorData(void);

		inline const std::string &getName(void) const { return _name; }
		inline void setName(const std::string &name) { _name = name; }

		inline uint getTitleHeight(void) const { return _title_height; }
		inline uint getPad(DirectionType dir) const { return _pad[(dir != DIRECTION_NO) ? dir : 0]; }

		// title textures
		inline PTexture *getTextureMain(FocusedState state) { return (state < FOCUSED_STATE_FOCUSED_SELECTED) ? _texture_main[state] : NULL; }
		inline PTexture *getTextureTab(FocusedState state) { return (state != FOCUSED_STATE_NO) ? _texture_tab[state] : NULL; }
		inline PTexture *getTextureSeparator(FocusedState state) { return (state < FOCUSED_STATE_FOCUSED_SELECTED) ? _texture_separator[state] : NULL; }

		// font
		inline PFont *getFont(FocusedState state) {
			if (state != FOCUSED_STATE_NO) {
				return (_font[state] == NULL) ? _font[FOCUSED_STATE_FOCUSED] : _font[state];
			}
			return NULL;
		}
		inline PFont::Color *getFontColor(FocusedState state) { return (state != FOCUSED_STATE_NO) ? _font_color[state] : NULL; }

		// border
		inline PTexture *getBorderTexture(FocusedState state, BorderPosition pos) {
			return (state < FOCUSED_STATE_FOCUSED_SELECTED) ? _texture_border[state][pos] : NULL; }

		// button
		inline std::list<Theme::PDecorButtonData*>::iterator buttonBegin(void) {
			return _button_list.begin(); }
		inline std::list<Theme::PDecorButtonData*>::iterator buttonEnd(void) {
			return _button_list.end(); }

		bool load(CfgParser::Section *cs);
		void unload(void);

		// sanity functions
		void check(void);

	private:
		void loadBorder(CfgParser::Section *cs);
		void loadButtons(CfgParser::Section *cs);

		void checkTextures(void);
		void checkFonts(void);
		void checkBorder(void);
		void checkColors(void);

	private:
		std::string _name;

		// size, padding etc
		uint _title_height;
		uint _pad[DIRECTION_NO];

		// title
		PTexture *_texture_main[FOCUSED_STATE_FOCUSED_SELECTED];
		PTexture *_texture_tab[FOCUSED_STATE_NO];
		PTexture *_texture_separator[FOCUSED_STATE_FOCUSED_SELECTED];

		// font
		PFont *_font[FOCUSED_STATE_NO];
		PFont::Color *_font_color[FOCUSED_STATE_NO];

		// border
		PTexture *_texture_border[FOCUSED_STATE_FOCUSED_SELECTED][BORDER_NO_POS];

		// buttons
		std::list<Theme::PDecorButtonData*> _button_list;

		static std::map<FocusedState, std::string> _fs_map;
		static std::map<BorderPosition, std::string> _border_map;
	};

	class PMenuData {
	public:
		PMenuData(void);
		~PMenuData(void);

		inline PFont *getFont(ObjectState state) { return _font[state]; }
		inline PFont::Color *getColor(ObjectState state) { return _color[state]; }
		inline PTexture *getTextureMenu(ObjectState state) { return _tex_menu[state]; }
		inline PTexture *getTextureItem(ObjectState state) { return _tex_item[state]; }
		inline PTexture *getTextureArrow(ObjectState state) { return _tex_arrow[state]; }
		inline PTexture *getTextureSeparator(ObjectState state) {
			return (state < OBJECT_STATE_SELECTED) ? _tex_sep[state] : _tex_sep[OBJECT_STATE_FOCUSED]; }
		inline uint getPad(DirectionType dir) const { return _pad[(dir != DIRECTION_NO) ? dir : 0]; }

		void load(CfgParser::Section *cs);
		void unload(void);

		// sanity functions
		void check(void);

	private:
		void loadState(CfgParser::Section *cs, ObjectState state);

	private:
		PFont *_font[OBJECT_STATE_NO + 1];
		PFont::Color *_color[OBJECT_STATE_NO + 1];
		PTexture *_tex_menu[OBJECT_STATE_NO + 1];
		PTexture *_tex_item[OBJECT_STATE_NO + 1];
		PTexture *_tex_arrow[OBJECT_STATE_NO + 1];
		PTexture *_tex_sep[OBJECT_STATE_NO];

		uint _pad[DIRECTION_NO];
	};

	class TextDialogData {
	public:
		TextDialogData(void);
		~TextDialogData(void);

		inline PFont *getFont(void) { return _font; }
		inline PFont::Color *getColor(void) { return _color; }
		inline PTexture *getTexture(void) { return _tex; }

		inline uint getPad(DirectionType dir) const { return _pad[(dir != DIRECTION_NO) ? dir : 0]; }

		void load(CfgParser::Section *cs);
		void unload(void);

		// sanity functions
		void check(void);

	private:
		PFont *_font;
		PFont::Color *_color;
		PTexture *_tex;

		uint _pad[DIRECTION_NO];
	};

	Theme(PScreen *scr);
	~Theme(void);

	void load(const std::string &dir);
	void unload(void);

	inline const GC &getInvertGC(void) const { return _invert_gc; }

	inline std::map<std::string, Theme::PDecorData*>::const_iterator decor_begin(void) { return _pdecordata_map.begin(); }
	inline std::map<std::string, Theme::PDecorData*>::const_iterator decor_end(void) { return _pdecordata_map.end(); }

	Theme::PDecorData *getPDecorData(const std::string &name) {
		std::map<std::string, Theme::PDecorData*>::iterator it = _pdecordata_map.begin();
		for (; it != _pdecordata_map.end(); ++it) {
			if (strcasecmp(it->first.c_str(), name.c_str()) == 0) {
				return it->second;
			}
		}
		return NULL;
	}

	// menu
	inline Theme::PMenuData *getMenuData(void) { return &_menu_data; }

#ifdef HARBOUR
	inline PTexture *getHarbourTexture(void) const { return _harbour_texture; }
#endif // HARBOUR

	// status/cmd
	inline Theme::TextDialogData *getStatusData(void) { return &_status_data; }
	inline Theme::TextDialogData *getCmdDialogData(void) { return &_cmd_d_data; }

private:
	PScreen *_scr;
	ImageHandler *_image_handler;
	std::string _theme_dir;

	bool _is_loaded;

	// gc
	GC _invert_gc;

	// frame decors
	std::map<std::string, Theme::PDecorData*> _pdecordata_map;

	// menu
	Theme::PMenuData _menu_data;

	// harbour
#ifdef HARBOUR
	PTexture *_harbour_texture;
#endif // HARBOUR

	// status window
	TextDialogData _status_data, _cmd_d_data;
};

#endif // _THEME_HH_
